package sockprog.example6_thread;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketTimeoutException;

public final class MessageClient {
    private static final int DEFAULT_BUFFER_SIZE = 32 * 1024;
    private static final int NO_TIMEOUT = 0;

    /**
     * クライアントプログラムを起動する。
     * 
     * <pre>
     * java sockprog.example6_thread.MessageClient \
     * [-t タイムアウト] [-b バッファサイズ] サーバーアドレス ポート番号 送信文字列
     * </pre>
     * 
     * のように起動する。 上記で[]で括られたものはオプションである。
     */
    public static void main(String[] args) {
        int timeout = NO_TIMEOUT;
        int bufferSize = DEFAULT_BUFFER_SIZE;

        int offset = 0;
        for(offset = 0; offset < args.length;) {
            if(args.length >= offset + 2 && args[offset].equals("-t")) {
                timeout = Integer.parseInt(args[offset + 1]) * 1000;
                offset += 2;
                continue;
            }
            if(args.length >= offset + 2 && args[offset].equals("-b")) {
                bufferSize = Integer.parseInt(args[offset + 1]) * 1024;
                offset += 2;
                continue;
            }
            if(args.length >= offset + 1 && args[offset].startsWith("-")) {
                System.err.println("unknown option: " + args[offset]);
                System.exit(-1);
            }
            break;
        }

        if(args.length < offset+3) {
            System.err.println("parameter not enough");
            System.exit(-1);
        }

        String host = args[offset];
        int port = Integer.parseInt(args[offset+1]);
        String message = args[offset+2];

        MessageClient client = null;
        try {
            client = new MessageClient(host, port, timeout, bufferSize);
            String reply = client.send(message);
            System.out.println(reply);
        } catch (SocketTimeoutException e) {
            // タイムアウトの際の処理を行う
            System.err.println("timeout");
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                if(client != null) {
                    client.shutDown();
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    private final Socket socket;
    private final Writer writer;
    private final BufferedReader reader;

    public MessageClient(String host, int port,
            int timeout, int bufferSize) throws IOException {
        socket = new Socket();

        // 送受信バッファサイズの設定
        socket.setReceiveBufferSize(bufferSize);
        socket.setSendBufferSize(bufferSize);

        // タイムアウトの設定
        socket.setSoTimeout(timeout);

        socket.connect(new InetSocketAddress(host, port));

        // writerの準備
        OutputStream out = socket.getOutputStream();
        OutputStreamWriter outWriter =
            new OutputStreamWriter(out, ProtocolConstants.ENCODING);
        writer = new BufferedWriter(outWriter, bufferSize);

        // readerの準備
        InputStream in = socket.getInputStream();
        InputStreamReader inReader =
            new InputStreamReader(in, ProtocolConstants.ENCODING);
        reader = new BufferedReader(inReader, bufferSize);
    }
    public void shutDown() throws IOException {
        socket.close();
    }
    public String send(String message) throws IOException {
        // 文字列の送信
        writer.write(message);
        writer.write(ProtocolConstants.LINE_SEPARATOR);
        writer.flush();

        // 文字列の受信
        return reader.readLine();
    }
}
