#!/usr/bin/ruby
# -*- coding: utf-8 -*-
# ↑Emacsで文字コードを強制するときのタグです。

### 受信用スクリプトです。なお、Cygwinではcygwin1.dllの関係でIPv6が使えません。

require 'ipaddr'
require 'socket'
require 'monitor'
require 'mcconfig'

is_ipv6 = ARGV[0] == '6'

if is_ipv6 && !defined?(Socket::IPV6_JOIN_GROUP)
  # IPv6有効時、Socket::IPV6_JOIN_GROUPが未定義ならば、Linuxでしか動かせま
  # せん。最新のソースからビルドした処理系ならば、IPv6関係の定数は定義済み
  # なので、この部分は無視されます。
  if /linux/ =~ RUBY_PLATFORM
    Socket::IPPROTO_IPV6 = 41
    Socket::IPV6_JOIN_GROUP = 20
  else
    raise "#{RUBY_PLATFORM} is not supported."
  end
end

# モニタ機能を用いて、スレッド間での同期を行います。
# 詳細はMonitorMixinのドキュメントを参照のこと。
msgs = Hash.new.extend(MonitorMixin)

# 受信処理と表示処理で扱うメッセージを定義します。
class Message
  attr_reader :cnt
  attr_reader :name
  attr_reader :addr
  attr_reader :port
  attr_reader :reg_time
  def initialize(cnt, name, addr, port)
    @cnt = cnt
    @name = name
    @addr = addr
    @port = port
    @reg_time = Time.now
  end
end

# 標準出力をノンバッファリングモードに指定します。
$stdout.sync = true

# ・引数でIPv4/IPv6を切り替えます。
# ・マルチキャスト受信用のソケットの設定です。Pythonとほとんど同じです。
if is_ipv6
  puts "Mode: IPv6\n"
  msock = UDPSocket.open(Socket::AF_INET6)
  msock.setsockopt(Socket::SOL_SOCKET, Socket::SO_REUSEADDR, true)
  msock.bind("::0", MULTICAST_PORT)
  mreq = IPAddr.new(MULTICAST_ADDR6).hton + [0].pack("I")
  msock.setsockopt(Socket::IPPROTO_IPV6, Socket::IPV6_JOIN_GROUP, mreq)
else
  puts "Mode: IPv4\n"
  msock = UDPSocket.open(Socket::AF_INET)
  msock.setsockopt(Socket::SOL_SOCKET, Socket::SO_REUSEADDR, true)
  msock.bind("0.0.0.0", MULTICAST_PORT)
  mreq = IPAddr.new(MULTICAST_ADDR4).hton + IPAddr.new("0.0.0.0").hton
  msock.setsockopt(Socket::IPPROTO_IP, Socket::IP_ADD_MEMBERSHIP, mreq)
end

# 表示処理を別スレッドで起動します。処理をブロック内に閉じた形で記述
# するため、メソッド定義にはしていません。
Thread.new {
  loop {
    now = Time.now
    expire_time = now - (DISP_EXPIRES / 1000.0)
    puts now.strftime("Ruby [%Y-%m-%d %H:%M:%S]")
    # 排他ロックしてデータを読み書きします。
    msgs.synchronize {
      msgs.values.sort {|a, b| a.name <=> b.name}.each {|msg|
        if msg.reg_time < expire_time
          msgs.delete(msg.name)
        else
          puts "  %04d %s (%s/%d)" % [msg.cnt, msg.name, msg.addr, msg.port]
        end
      }
    }
    puts
    sleep(DISP_INTERVAL / 1000.0) # ミリ秒→秒
  }
}

# 受信処理は現在のスレッドで行います。
loop {
  data, saddr = msock.recvfrom(RECV_BUF_SIZE)
  args = data.split(/\s+/, 2) + [saddr[3], saddr[1]]
  msg = Message.new(*args)
  # 排他ロックしてデータを書き込みます。
  msgs.synchronize { msgs[msg.name] = msg }
}
