#!/usr/bin/python -u
# -*- coding: utf-8 -*-
# ↑もともとはEmacsで文字コードを強制するときのタグですが、Python自身がファ
# イルの文字コードを判定するためにも使います。判定した文字コードと実際の文
# 字コードが一致していない場合、たとえコメント中であってもエラーになります。
#
# コマンド引数の「-u」は標準出力をノンバッファリングモードに指定しています。

### 受信用スクリプトです。なお、Cygwinではcygwin1.dllの関係でIPv6が使えません。

import sys
from socket import * # 本来'*'でインポートすることは推奨されないようです。
from struct import pack
from time import time, localtime, strftime, sleep
from thread import start_new_thread, allocate_lock
from mcconfig import *

# inet_ptonが未定義の場合、簡易版のinet_ptonを定義します。
if 'inet_pton' not in globals():
    def inet_pton(family, addr):
        if family == AF_INET:
            return inet_aton(addr)
        elif family == AF_INET6:
            alist = [0] * 8
            hl = addr.split("::", 2)
            if len(hl) == 1:
                hl += [""]
            h, l = hl
            if len(h) > 0:
                i = 0
                for e in h.split(":"):
                    alist[i] = int(e, 16)
                    i += 1
            if len(l) > 0:
                i = 7
                for e in reversed(l.split(":")):
                    alist[i] = int(e, 16)
                    i -= 1
            return pack("!8H", *alist)

# スレッド間での同期のため、ロックを確保します。
# 他の多くの言語と違い、ロックオブジェクトを明示的に扱う必要があるようです。
msgs = {}
msgs_lock = allocate_lock()

# 受信処理と表示処理で扱うメッセージを定義します。
class Message:
    def __init__(self, cnt, name, addr, port):
        self.cnt = int(cnt)
        self.name = name
        self.addr = addr
        self.port = port
        self.reg_time = time()

# ・引数でIPv4/IPv6を切り替えます。
# ・マルチキャスト受信用のソケットの設定です。Rubyとほとんど同じです。
if len(sys.argv) > 1 and sys.argv[1] == '6':
    msock = socket(AF_INET6, SOCK_DGRAM)
    msock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    msock.bind(("::0", MULTICAST_PORT))
    mreq = inet_pton(AF_INET6, MULTICAST_ADDR6) + pack("I", 0)
    msock.setsockopt(IPPROTO_IPV6, IPV6_JOIN_GROUP, mreq)
else:
    msock = socket(AF_INET, SOCK_DGRAM)
    msock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    msock.bind(("0.0.0.0", MULTICAST_PORT))
    mreq = inet_pton(AF_INET, MULTICAST_ADDR4) + inet_pton(AF_INET, "0.0.0.0")
    msock.setsockopt(IPPROTO_IP, IP_ADD_MEMBERSHIP, mreq)

# 表示処理です。
def display():
    while True:
        now = time()
        expire_time = now - (DISP_EXPIRES / 1000.0)
        print strftime("Python [%Y-%m-%d %H:%M:%S]", localtime(now))
        # 排他ロックしてデータを読み書きします。
        msgs_lock.acquire()
        try:
            values = msgs.values()
            values.sort(lambda a, b: cmp(a.name, b.name))
            for msg in values:
                if msg.reg_time < expire_time:
                    del msgs[msg.name]
                else:
                    print "  %04d %s (%s/%d)" % \
                        (msg.cnt, msg.name, msg.addr, msg.port)
        finally:
            msgs_lock.release() # 明示的にロックを解除する必要があります。
        print
        sleep(DISP_INTERVAL / 1000.0) # ミリ秒→秒

# 表示処理を別スレッドで起動します。
start_new_thread(display, ())

# 受信処理は現在のスレッドで行います。
while True:
    data, saddr = msock.recvfrom(RECV_BUF_SIZE)
    cnt, name = data.split(None, 2)
    msg = Message(cnt, name, saddr[0], saddr[1])
    # 排他ロックしてデータを書き込みます。
    msgs_lock.acquire()
    try:
        msgs[msg.name] = msg
    finally:
        msgs_lock.release() # 明示的にロックを解除する必要があります。
